/*************************************************************************\
* Copyright (c) 2016 UChicago Argonne LLC, as Operator of Argonne
*     National Laboratory.
* Copyright (c) 2002 The Regents of the University of California, as
*     Operator of Los Alamos National Laboratory.
* EPICS BASE is distributed subject to a Software License Agreement found
* in file LICENSE that is included with this distribution.
\*************************************************************************/

/* devApsEr.c */

/*
 *      Author:         John R. Winans
 *      Date:           07-21-93
 *      Modified for R3.14/OSI, W. Eric Norum, 2005-01-20
 */

#include <epicsStdlib.h>
#include <epicsStdio.h>
#include <epicsThread.h>
#include <epicsMutex.h>
#include <epicsInterrupt.h>
#include <epicsExit.h>
#include <epicsExport.h>
#include <devLib.h>
#include <iocsh.h>

#include <devSup.h>
#include <recGbl.h>
#include <dbDefs.h>
#include <dbAccess.h>
#include <epicsPrint.h>
#include <link.h>
#include <dbScan.h>
#include <eventRecord.h>
#include <dbCommon.h>

#include <erDefs.h>
#include <erRecord.h>
#include <ereventRecord.h>
#include "drvTS.h"
#include "devApsEr.h"


typedef struct ApsErStruct
{
  epicsUInt16 Control;
  epicsUInt16 EventRamAddr;
  epicsUInt16 EventRamData;
  epicsUInt16 OutputPulseEnables;
  epicsUInt16 OutputLevelEnables;
  epicsUInt16 TriggerEventEnables;
  epicsUInt16 EventCounterLo;
  epicsUInt16 EventCounterHi;
  epicsUInt16 TimeStampLo;
  epicsUInt16 TimeStampHi;

  epicsUInt16 EventFifo;
  epicsUInt16 EventTimeHi;
  epicsUInt16 DelayPulseEnables;
  epicsUInt16 DelayPulseSelect;
  epicsUInt16 PulseDelay;
  epicsUInt16 PulseWidth;
  epicsUInt16 IrqVector;
  epicsUInt16 IrqEnables;
} ApsErStruct;


#define NUM_EVENTS      256

#define ER_CSR_ER_ENABLE    0x8000
#define ER_CSR_INT_ENABLE   0x4000
#define ER_CSR_TS_RESET     0x2000
#define ER_CSR_HEARTBEAT    0x1000
#define ER_CSR_EVENT_INT    0x0800
#define ER_CSR_TS_LATCH     0x0400
#define ER_CSR_MAP_ENABLE   0x0200
#define ER_CSR_MAP_SELECT   0x0100
#define ER_CSR_MAP_MASK     0x0300
#define ER_CSR_MAP_RAM1     (ER_CSR_MAP_ENABLE)
#define ER_CSR_MAP_RAM2     (ER_CSR_MAP_ENABLE | ER_CSR_MAP_SELECT)
#define ER_CSR_RAM_CLEAR    0x0080
#define ER_CSR_RAM_SELECT   0x0040
#define ER_CSR_RAMADR_AUTO  0x0020
#define ER_CSR_RAMADR_RESET 0x0010
#define ER_CSR_FIFO_RESET   0x0008
#define ER_CSR_FIFO_FULL    0x0004
#define ER_CSR_FIFO_DATA    0x0002
#define ER_CSR_TAXI_ERROR   0x0001

/*
 * These control/status bits are normally preserved and written back
 * to the control register unchanged when adjusting other bits
 */
#define ER_CSR_SAVE_MASK \
 (ER_CSR_ER_ENABLE | ER_CSR_INT_ENABLE | ER_CSR_MAP_MASK | \
  ER_CSR_RAM_SELECT | ER_CSR_RAMADR_AUTO)


#define ER_IRQ_ALL      0x000f
#define ER_IRQ_OFF      0x0000
#define ER_IRQ_TELL     0xffff
#define ER_IRQ_EVENT    0x0008
#define ER_IRQ_HEART    0x0004
#define ER_IRQ_FIFO     0x0002
#define ER_IRQ_TAXI     0x0001

typedef struct ErLinkStruct
{
  struct erRecord *pRec;
  int             Card;
  long            IrqVector;
  long            IrqLevel;
  ApsErStruct     *pEr;
  epicsMutexId    LinkLock;               /* Lock for the card */
  epicsUInt16     ErEventTab[NUM_EVENTS]; /* current view of the RAM */
  IOSCANPVT       IoScanPvt[NUM_EVENTS];  /* event-based rec processing */
  TSeventFunc     EventFunc;
  TSerrorFunc     ErrorFunc;
  int             faultInterruptsMasked;
} ErLinkStruct;

#define MAX_ER_CARDS 4

static int NumErCards = 0;
static int ConfigureLock = 0;
static ErLinkStruct ErLink[MAX_ER_CARDS];

int ErDebug = 0;
epicsExportAddress(int,ErDebug);


/* Forward definitions */

static long ErMasterEnableGet(ApsErStruct *pParm);
static long ErMasterEnableSet(ApsErStruct *pParm, int Enable);
static long ErSetTrg(ApsErStruct *pParm, int Channel, int Enable);
static long ErSetOtp(ApsErStruct *pParm, int Channel, int Enable);
static long ErSetOtl(ApsErStruct *pParm, int Channel, int Enable);
static long ErSetDg(ApsErStruct *pParm, int Channel, int Enable, epicsUInt16 Delay, epicsUInt16 Width);
static epicsUInt16 ErEnableIrq(ApsErStruct *pParm, epicsUInt16 Mask);
static int ErRegisterIrq(ErLinkStruct *pLink, int IrqVector, int IrqLevel);
static long ErCheckTaxi(ApsErStruct *pParm);
static int ErGetRamStatus(ApsErStruct *pParm, int Ram);
static int ErResetAll(ApsErStruct *pParm);
static int ErUpdateRam(ApsErStruct *pParm, epicsUInt16 *RamBuf);
static int ErProgramRam(ApsErStruct *pParm, epicsUInt16 *RamBuf, int Ram);
static int ErEnableRam(ApsErStruct *pParm, int Ram);
static long proc(struct erRecord  *pRec);

/******************************************************************************
 *
 * Routine used to verify that a given card number is valid.
 * Returns zero if valid, nonzero otherwise.
 *
 ******************************************************************************/
long ErHaveReceiver(int Card)
{
  if ((Card < 0) || (Card >= NumErCards))
    return -1;
  if (ErLink[Card].pEr == NULL)
    return -1;

  return NUM_EVENTS; /* Return the number of possible event codes, 0-255 */
}

/******************************************************************************
 *
 * Register a listener for the event system.  Every time we get an event from
 * the event receiver, we will call the registered listener and pass in the
 * event number received and the tick counter value when that event was
 * received.  If a different event listener was already registered we store
 * the new one and pass the old one to our caller; the new listener must then
 * call the old one.
 *
 ******************************************************************************/
TSeventFunc ErRegisterEventHandler(int Card, TSeventFunc func)
{
  TSeventFunc prev = ErLink[Card].EventFunc;

  if (ErDebug)
    printf("ErRegisterEventHandler(%d, %p)\n", Card, func);

  if (func == prev)
    return NULL;

  ErLink[Card].EventFunc = func;
  return prev;
}

/******************************************************************************
 *
 * Register an error handler for the event system.  When we get an error from
 * the event receiver, we will call the registered handler and pass in the
 * error number received.  If a different error handler was already registered
 * we store the new one and return the old one to our caller; the new error
 * handler must then call the old one.
 *
 ******************************************************************************/
TSerrorFunc ErRegisterErrorHandler(int Card, TSerrorFunc func)
{
  TSerrorFunc prev = ErLink[Card].ErrorFunc;

  if (ErDebug)
    printf("ErRegisterErrorHandler(%d, %p)\n", Card, func);

  if (func == prev)
    return NULL;

  ErLink[Card].ErrorFunc = func;
  return prev;
}

/******************************************************************************
 *
 * Return the current tick counter value.
 *
 ******************************************************************************/
long ErGetTicks(int Card, epicsUInt32 *Ticks)
{
  int key;
  volatile ApsErStruct *pEr;
  epicsUInt16 status;
  epicsUInt32 ticks;

  if (ErHaveReceiver(Card) < 0)
    return -1;

  pEr = ErLink[Card].pEr;

  key = epicsInterruptLock();
  pEr->Control = (pEr->Control & ER_CSR_SAVE_MASK) | ER_CSR_TS_LATCH;
  status = pEr->Control; /* Flush earlier writes */
  ticks = pEr->TimeStampLo | pEr->TimeStampHi << 16;
  epicsInterruptUnlock(key);

  *Ticks = ticks;

  return 0;
}


/******************************************************************************
 *
 * This routine is to be called in the startup script in order to init the
 * card addresses and the associated IRQ vectors and levels.
 *
 * By default there are no cards configured.
 *
 ******************************************************************************/
int ErConfigure(int Card, epicsUInt32 CardAddress, int IrqVector, int IrqLevel)
{
  short Junk;
  static int unregistered = 1;
  static struct drvTSops apsErOps = {
    ErRegisterEventHandler, /* TSregisterEventHandler */
    ErRegisterErrorHandler, /* TSregisterErrorHandler */
    ErGetTicks,             /* TSgetTicks */
    ErHaveReceiver,         /* TShaveReceiver */
    NULL,                   /* TSforceSync */
    NULL,                   /* TSgetTime */
    NULL,                   /* TSsyncEvent */
    NULL,                   /* TSdirectTime */
    NULL,                   /* TSdriverInit */
    NULL,                   /* TSuserGet */
  };

  if (ConfigureLock != 0) {
    epicsPrintf("devApsEr: Too late to change ER configuration\n");
    return -1;
  }

  if (Card >= MAX_ER_CARDS) {
    epicsPrintf("devApsEr: Can't drive more than %d cards\n", MAX_ER_CARDS);
    return -1;
  }

  if (CardAddress > 0xffff) {
    epicsPrintf("devApsEr: Card address invalid, must be 0x0000-0xffff\n");
    return -1;
  }

  if(devRegisterAddress("apsEr", atVMEA16, CardAddress, 0x40,
      (void *)&ErLink[Card].pEr) != 0) {
    epicsPrintf("devApsEr: Failed to map requested A16 address\n");
    ErLink[Card].pEr = NULL;
    return -1;
  }

  if (devReadProbe(sizeof(short), &ErLink[Card].pEr->Control, &Junk) != 0) {
    epicsPrintf("devApsEr: Probe for event receiver failed\n");
    ErLink[Card].pEr = NULL;
    return -1;
  }

  ErLink[Card].IrqVector = IrqVector;
  ErLink[Card].IrqLevel = IrqLevel;

  if (Card >= NumErCards)
    NumErCards = Card+1;

  if (unregistered) {
    drvTsRegisterDeviceSupport(&apsErOps);
    unregistered = 0;
  }
  return 0;
}

/******************************************************************************
 *
 * A function to shut down all event receivers when we get a ^X style reboot.
 *
 ******************************************************************************/
static void ErRebootFunc(void *arg)
{
  int j = 0;

  while (j < NumErCards)
  {
    if (ErLink[j].pEr != NULL)
      ErEnableIrq(ErLink[j].pEr, ER_IRQ_OFF);
    ++j;
  }
}

/******************************************************************************
 *
 * Register and init the IRQ handlers for each ER card.
 *
 ******************************************************************************/
static long ErInitDev(int pass)
{
  if (ErDebug)
    printf("ErInitDev(%d)\n", pass);

  if (!ConfigureLock) {
    int j;

    ConfigureLock = 1;  /* No more config/init allowed */

    epicsAtExit(ErRebootFunc, NULL);

    /* Clear the record pointers */
    for (j = 0; j < NumErCards; j++) {
      if (ErHaveReceiver(j) >= 0) {
        int k;

        ErLink[j].LinkLock = epicsMutexMustCreate();

        for (k = 0; k < NUM_EVENTS; k++) {
          ErLink[j].Card = j;
          ErLink[j].ErEventTab[k] = 0;
          scanIoInit(&ErLink[j].IoScanPvt[k]);
        }

        ErLink[j].pRec = NULL;
        ErResetAll(ErLink[j].pEr);
        ErRegisterIrq(&ErLink[j], ErLink[j].IrqVector, ErLink[j].IrqLevel);
      }
    }
  }

  if (pass == 1) {
    int j;

    for (j = 0; j < NumErCards; j++)
      if (ErHaveReceiver(j) >= 0)
        ErEnableIrq(ErLink[j].pEr, ER_IRQ_ALL);
  }
  return 0;
}

/******************************************************************************
 *
 * Init routine for the ER record type.
 *
 ******************************************************************************/
static long init_record(struct erRecord *pRec)
{
  ErLinkStruct  *pLink;

  if (ErDebug)
    printf("devApsEr::init_record() entered\n");

  if (ErHaveReceiver(pRec->out.value.vmeio.card) < 0) {
    recGblRecordError(S_db_badField, (void *)pRec,
        "devApsEr::ErInitErRec() Invalid card number");
    return S_db_badField;
  }

  pLink = &ErLink[pRec->out.value.vmeio.card];

  /* Make sure we only have one record for a given ER card */
  if (pLink->pRec != NULL) {
    recGblRecordError(S_db_badField, (void *)pRec,
        "devApsEr::ErInitErRec() only one erRecord allowed per card");
    return S_db_badField;
  }

  pLink->pRec = pRec;
  proc(pRec);

  return 0;
}

/******************************************************************************
 *
 * Process routine for the ER record type.
 *
 ******************************************************************************/
static long proc(struct erRecord  *pRec)
{
  ErLinkStruct  *pLink = &ErLink[pRec->out.value.vmeio.card];

  if (pRec->tpro > 10)
    printf("devApsEr::proc(%s) entered\n", pRec->name);

  /* Make sure the card is present */
  if (ErHaveReceiver(pRec->out.value.vmeio.card) < 0)
    return 0;

  epicsMutexLock(pLink->LinkLock);

  if (pRec->enab != ErMasterEnableGet(pLink->pEr))
    ErMasterEnableSet(pLink->pEr, pRec->enab);

  ErSetTrg(pLink->pEr, 0, pRec->trg0);
  ErSetTrg(pLink->pEr, 1, pRec->trg1);
  ErSetTrg(pLink->pEr, 2, pRec->trg2);
  ErSetTrg(pLink->pEr, 3, pRec->trg3);
  ErSetTrg(pLink->pEr, 4, pRec->trg4);
  ErSetTrg(pLink->pEr, 5, pRec->trg5);
  ErSetTrg(pLink->pEr, 6, pRec->trg6);

  ErSetOtp(pLink->pEr, 0, pRec->otp0);
  ErSetOtp(pLink->pEr, 1, pRec->otp1);
  ErSetOtp(pLink->pEr, 2, pRec->otp2);
  ErSetOtp(pLink->pEr, 3, pRec->otp3);
  ErSetOtp(pLink->pEr, 4, pRec->otp4);
  ErSetOtp(pLink->pEr, 5, pRec->otp5);
  ErSetOtp(pLink->pEr, 6, pRec->otp6);
  ErSetOtp(pLink->pEr, 7, pRec->otp7);
  ErSetOtp(pLink->pEr, 8, pRec->otp8);
  ErSetOtp(pLink->pEr, 9, pRec->otp9);
  ErSetOtp(pLink->pEr, 10, pRec->otpa);
  ErSetOtp(pLink->pEr, 11, pRec->otpb);
  ErSetOtp(pLink->pEr, 12, pRec->otpc);
  ErSetOtp(pLink->pEr, 13, pRec->otpd);

  ErSetOtl(pLink->pEr, 0, pRec->otl0);
  ErSetOtl(pLink->pEr, 1, pRec->otl1);
  ErSetOtl(pLink->pEr, 2, pRec->otl2);
  ErSetOtl(pLink->pEr, 3, pRec->otl3);
  ErSetOtl(pLink->pEr, 4, pRec->otl4);
  ErSetOtl(pLink->pEr, 5, pRec->otl5);
  ErSetOtl(pLink->pEr, 6, pRec->otl6);

  ErSetDg(pLink->pEr, 0, pRec->dg0e, pRec->dg0d, pRec->dg0w);
  ErSetDg(pLink->pEr, 1, pRec->dg1e, pRec->dg1d, pRec->dg1w);
  ErSetDg(pLink->pEr, 2, pRec->dg2e, pRec->dg2d, pRec->dg2w);
  ErSetDg(pLink->pEr, 3, pRec->dg3e, pRec->dg3d, pRec->dg3w);

  if (ErCheckTaxi(pLink->pEr) != 0)
    pRec->taxi = 1;
  else
    pRec->taxi = 0;

  epicsMutexUnlock(pLink->LinkLock);

  pRec->udf = 0;
  return 0;
}

static ErDsetStruct devEr = {
  5, NULL, ErInitDev, init_record, NULL, proc
};
epicsExportAddress(dset, devEr);

/******************************************************************************
 *
 * Init_Record routine for the EREVENT record type.
 *
 ******************************************************************************/
static long ErEventInitRec(struct ereventRecord *pRec)
{
  if (ErDebug)
    printf("ErEventInitRec(%s)\n", pRec->name);

  if (ErHaveReceiver(pRec->out.value.vmeio.card) < 0) {
    recGblRecordError(S_db_badField, (void *)pRec,
        "devApsEr::ErEventInitRec() invalid card number in INP field");
    return S_db_badField;
  }

  if ((pRec->out.value.vmeio.signal < 0) ||
      (pRec->out.value.vmeio.signal > 255)) {
    recGblRecordError(S_db_badField, (void *)pRec,
        "devApsEr::ErEventInitRec() invalid signal number in INP field");
    return S_db_badField;
  }

  pRec->lenm = 300;     /* Force setting on first process */
  pRec->lout = 0;       /* Clear the 'last' event mask */
  return 0;
}

/******************************************************************************
 *
 * Process routine for the EREVENT record type.
 *
 ******************************************************************************/
static long ErEventProc(struct ereventRecord  *pRec)
{
  ErLinkStruct  *pLink = &ErLink[pRec->out.value.vmeio.card];
  int           DownLoadRam = 0;
  unsigned      short Mask;

  if (pRec->tpro > 10)
    printf("ErEventProc(%s) entered\n", pRec->name);

  /* Make sure the card is present */
  if (ErHaveReceiver(pRec->out.value.vmeio.card) < 0)
    return 0;

  if (pRec->enab != 0) {
    if (pRec->tpro > 10)
      printf("ErEventProc(%s) enable=true\n", pRec->name);

    epicsMutexLock(pLink->LinkLock);
    if (pRec->enm != pRec->lenm) {
      if (pRec->tpro > 10)
        printf("ErEventProc(%s) event number changed %ld-%ld\n", pRec->name, (long)pRec->lenm, (long)pRec->enm);

      /* Transfer the event info to new position */
      if ((pRec->lenm < 256) && (pRec->lenm >= 0))
        pLink->ErEventTab[pRec->lenm] = 0;

      DownLoadRam = 1;
      pRec->lenm = pRec->enm;
    }

    /* Build event mask */
    /* NOTE: that we never latch the time in the 32 bit latch */

    Mask = 0;
    if (pRec->out0 != 0) Mask |= 0x0001;
    if (pRec->out1 != 0) Mask |= 0x0002;
    if (pRec->out2 != 0) Mask |= 0x0004;
    if (pRec->out3 != 0) Mask |= 0x0008;
    if (pRec->out4 != 0) Mask |= 0x0010;
    if (pRec->out5 != 0) Mask |= 0x0020;
    if (pRec->out6 != 0) Mask |= 0x0040;
    if (pRec->out7 != 0) Mask |= 0x0080;
    if (pRec->out8 != 0) Mask |= 0x0100;
    if (pRec->out9 != 0) Mask |= 0x0200;
    if (pRec->outa != 0) Mask |= 0x0400;
    if (pRec->outb != 0) Mask |= 0x0800;
    if (pRec->outc != 0) Mask |= 0x1000;
    if (pRec->outd != 0) Mask |= 0x2000;
    if (pRec->vme != 0)  Mask |= 0x8000;

    if (pRec->tpro > 10)
      printf("ErEventProc(%s) New RAM mask is 0x%4.4X\n", pRec->name, Mask);

    if (Mask != pRec->lout)
      DownLoadRam = 1;

    if (DownLoadRam != 0) {
      if (pRec->enm != 0) {
        pLink->ErEventTab[pRec->enm] = Mask;
        pRec->lout = Mask;
      }
      ErUpdateRam(pLink->pEr, pLink->ErEventTab);
    }

    if (pRec->tpro > 10)
      printf("ErEventProc(%s) I/O operations complete\n", pRec->name);

    epicsMutexUnlock(pLink->LinkLock);
  }
  else {
    /* BUG kill event if was ever enabled! */
  }

  return 0;
}

static ErDsetStruct devErevent = {
  5, NULL, NULL, ErEventInitRec, NULL, ErEventProc
};
epicsExportAddress(dset, devErevent);

/******************************************************************************
 *
 * Routine to initialize an event record.  This is a regular EPICS event
 * record!! Not an ER or EG event record.  We allow the user to use the
 * regular epics event records to request processing on an event that is
 * received from the event receiver.
 *
 * All we gotta do here is set the dpvt field to point to the proper
 * IOSCANEVENT field so that we can later return it on getIoScanInfo() calls.
 *
 ******************************************************************************/
static long ErEpicsEventInit(struct eventRecord *pRec)
{
  pRec->dpvt = NULL;    /* In case problems arise below */

  if(ErDebug)
    printf("ErEpicsEventInit(%s) %d %d\n",
      pRec->name, pRec->inp.value.vmeio.card, pRec->inp.value.vmeio.signal);

  if (ErHaveReceiver(pRec->inp.value.vmeio.card) < 0) {
    recGblRecordError(S_db_badField, (void *)pRec,
        "devApsEr::ErEventInitRec() invalid card number in INP field");
    return S_db_badField;
  }

  if ((pRec->inp.value.vmeio.signal < 0) ||
      (pRec->inp.value.vmeio.signal > 255)) {
    recGblRecordError(S_db_badField, (void *)pRec,
        "devApsEr::ErEventInitRec() invalid signal number in INP field");
    return S_db_badField;
  }

  pRec->dpvt = &ErLink[pRec->inp.value.vmeio.card].IoScanPvt[pRec->inp.value.vmeio.signal];
  return 0;
}

/******************************************************************************
 *
 * Routine to get the IOSCANPVT value for an EVENT record type.
 *
 ******************************************************************************/
static long ErEpicsEventGetIoScan(int cmd, struct eventRecord *pRec, IOSCANPVT *pPvt)
{
  if (pRec->dpvt) {
      *pPvt = *((IOSCANPVT*)(pRec->dpvt));
      return 0;
  }
  return TRUE; /* No hardware */
}

static ErDsetStruct devErEpicsEvent = {
  5, NULL, NULL, ErEpicsEventInit, ErEpicsEventGetIoScan, NULL
};
epicsExportAddress(dset, devErEpicsEvent);

/******************************************************************************
 *
 * Inquire if the card is enabled.
 *
 ******************************************************************************/
static long ErMasterEnableGet(ApsErStruct *pParm)
{
  volatile ApsErStruct *pEr = pParm;

  return (pEr->Control & ER_CSR_ER_ENABLE) != 0;
}

/******************************************************************************
 *
 * Enable/disable the card.
 *
 ******************************************************************************/
static long ErMasterEnableSet(ApsErStruct *pParm, int Enable)
{
  volatile ApsErStruct *pEr = pParm;

  if (Enable != 0)
    pEr->Control = (pEr->Control & ER_CSR_SAVE_MASK) | ER_CSR_ER_ENABLE;
  else
    pEr->Control = (pEr->Control & ER_CSR_SAVE_MASK) & ~ER_CSR_ER_ENABLE;
  return 0;
}

/******************************************************************************
 *
 * Check to see if we have a taxi violation and reset it if so.
 *
 ******************************************************************************/
static long ErCheckTaxi(ApsErStruct *pParm)
{
  volatile ApsErStruct *pEr = pParm;

  if (pEr->Control & ER_CSR_TAXI_ERROR) {
    pEr->Control = (pEr->Control & ER_CSR_SAVE_MASK) | ER_CSR_TAXI_ERROR;
    return 1;
  }
  return 0;
}

/******************************************************************************
 *
 * Set/clear an enable bit in the trigger mask.
 *
 ******************************************************************************/
static long ErSetTrg(ApsErStruct *pParm, int Channel, int Enable)
{
  volatile ApsErStruct *pEr = pParm;
  epicsUInt16 mask = 1 << Channel;

  if (Enable != 0)
    pEr->TriggerEventEnables |= mask;
  else
    pEr->TriggerEventEnables &= ~mask;
  return 0;
}
/******************************************************************************
 *
 * Set/clear an enable bit in the one-shot mask.
 *
 ******************************************************************************/
static long ErSetOtp(ApsErStruct *pParm, int Channel, int Enable)
{
  volatile ApsErStruct *pEr = pParm;
  epicsUInt16 mask = 1 << Channel;

  if (Enable != 0)
    pEr->OutputPulseEnables |= mask;
  else
    pEr->OutputPulseEnables &= ~mask;
  return 0;
}
/******************************************************************************
 *
 * Set/clear an enable bit in the level-trigger mask.
 *
 ******************************************************************************/
static long ErSetOtl(ApsErStruct *pParm, int Channel, int Enable)
{
  volatile ApsErStruct *pEr = pParm;
  epicsUInt16 mask = 1 << Channel;

  if (Enable != 0)
    pEr->OutputLevelEnables |= mask;
  else
    pEr->OutputLevelEnables &= ~mask;
  return 0;
}
/******************************************************************************
 *
 * Set/clear an enable bit in the programmable pulse delay mask.  If enabling
 * a channel, the delay and width are also set.
 *
 ******************************************************************************/
static long ErSetDg(ApsErStruct *pParm, int Channel, int Enable, epicsUInt16 Delay, epicsUInt16 Width)
{
  volatile ApsErStruct *pEr = pParm;
  epicsUInt16 mask = 1 << Channel;

  if (Enable != 0) {
    pEr->DelayPulseSelect = Channel;
    pEr->PulseDelay = Delay;
    pEr->PulseWidth = Width;
    pEr->DelayPulseEnables |= mask;
  }
  else
    pEr->DelayPulseEnables &= ~mask;
  return 0;
}

/******************************************************************************
 *
 * A debugging aid that dumps out the regs on the ER board.
 *
 ******************************************************************************/
static int ErDumpRegs(ApsErStruct *pParm)
{
  volatile ApsErStruct *pEr = pParm;

  printf("  Control/Status = %4.4X\n"
         "  EventRamAddr = %2.2X\n"
         "  EventRamData = %4.4X\n"
         "  OutputPulseEnables = %4.4X\n"
         "  OutputLevelEnables = %4.4X\n"
         "  TriggerEventEnables = %2.2X\n"
         "  EventCounter = %4.4X  %4.4X\n"
         "  TimeStamp = %4.4X %4.4X\n",
        pEr->Control, pEr->EventRamAddr & 0xFF, pEr->EventRamData,
        pEr->OutputPulseEnables, pEr->OutputLevelEnables,
        pEr->TriggerEventEnables & 0x7F, pEr->EventCounterHi,
        pEr->EventCounterLo, pEr->TimeStampHi, pEr->TimeStampLo);

  printf("  (FIFO registers excluded)\n"
         "  DelayPulseEnables = %1.1X\n"
         "  DelayPulseSelect = %1.1X\n"
         "  PulseDelay = %4.4X\n"
         "  PulseWidth = %4.4X\n"
         "  IrqVector = %4.4X\n"
         "  IrqEnables = %1.1X\n",
        pEr->DelayPulseEnables & 0xF, pEr->DelayPulseSelect & 0x3,
        pEr->PulseDelay, pEr->PulseWidth, pEr->IrqVector,
        pEr->IrqEnables & 0xF);

  return 0;
}

/******************************************************************************
 *
 * Reset, clear and disable everything on the board.
 *
 ******************************************************************************/
static int ErResetAll(ApsErStruct *pParm)
{
  volatile ApsErStruct *pEr = pParm;

  pEr->Control = (ER_CSR_TS_RESET | ER_CSR_RAMADR_RESET | ER_CSR_FIFO_RESET |
    ER_CSR_FIFO_FULL | ER_CSR_TAXI_ERROR);
  pEr->Control = (ER_CSR_TS_RESET | ER_CSR_RAM_CLEAR | ER_CSR_FIFO_RESET |
    ER_CSR_FIFO_FULL | ER_CSR_TAXI_ERROR);

  while (pEr->Control & ER_CSR_RAM_CLEAR)
    epicsThreadSleep(0.01);

  pEr->OutputPulseEnables = 0;
  pEr->OutputLevelEnables = 0;
  pEr->DelayPulseEnables = 0;
  pEr->DelayPulseSelect = 0;
  pEr->IrqEnables = 0;

  pEr->Control = (pEr->Control & ER_CSR_SAVE_MASK) | ER_CSR_ER_ENABLE;

  return 0;
}

/******************************************************************************
 *
 * Receive a hardware IRQ from an ER board.
 *
 * BUG -- should add errMessage calls for problem events & include storm
 *        control on them.
 *
 ******************************************************************************/
void ErIrqHandler(void *pvt)
{
  ErLinkStruct *pLink = (ErLinkStruct *)pvt;
  volatile ApsErStruct *pEr = pLink->pEr;
  epicsUInt16 status = pEr->Control;

  if (!(status & (ER_CSR_HEARTBEAT | ER_CSR_FIFO_FULL | ER_CSR_TAXI_ERROR))) {
    if (pLink->faultInterruptsMasked) {
        pEr->IrqEnables |= pLink->faultInterruptsMasked;
        pLink->faultInterruptsMasked = 0;
    }
  }
  else {
    if (status & ER_CSR_HEARTBEAT)
    { /* Got a lost heart beat */
      pEr->IrqEnables &= ~ER_IRQ_HEART;
      pEr->Control = (status & ER_CSR_SAVE_MASK) | ER_CSR_HEARTBEAT;
      pLink->faultInterruptsMasked |= ER_IRQ_HEART;

      if (pLink->ErrorFunc != NULL)
        (*pLink->ErrorFunc)(pLink->Card, ERROR_HEART);

      if(ErDebug)
        epicsInterruptContextMessage("ErIrqHandler() lost heart beat\n");
    }
    /* Kill any fifo overflow status */
    if (status & ER_CSR_FIFO_FULL)
    {
      pEr->IrqEnables &= ~ER_IRQ_FIFO;
      pEr->Control = (status & ER_CSR_SAVE_MASK) | ER_CSR_FIFO_FULL;
      pLink->faultInterruptsMasked |= ER_IRQ_FIFO;

      if (pLink->ErrorFunc != NULL)
        (*pLink->ErrorFunc)(pLink->Card, ERROR_LOST);

      if(ErDebug)
          epicsInterruptContextMessage("ErIrqHandler() FIFO overflow\n");
    }
    if (status & ER_CSR_TAXI_ERROR)
    {
      pEr->IrqEnables &= ~ER_IRQ_TAXI;
      pEr->Control = (status & ER_CSR_SAVE_MASK) | ER_CSR_TAXI_ERROR;
      pLink->faultInterruptsMasked |= ER_IRQ_TAXI;

      if (pLink->ErrorFunc != NULL)
        (*pLink->ErrorFunc)(pLink->Card, ERROR_TAXI);

      if(ErDebug)
        epicsInterruptContextMessage("ErIrqHandler() taxi violation\n");
    }
  }

  /* Process events waiting in the FIFO */
  if (status & ER_CSR_EVENT_INT) {
    int limit = 10;

    pEr->Control = (status & ER_CSR_SAVE_MASK) | ER_CSR_EVENT_INT;

    while ((pEr->Control & ER_CSR_FIFO_DATA) && (--limit > 0)) {
      epicsUInt16 j, k, EventNum;
      epicsUInt32 Time;
      int key;

      /* De-queue one event */
      key = epicsInterruptLock();
      j = pEr->EventFifo;
      k = pEr->EventTimeHi;
      epicsInterruptUnlock(key);

      /* Extract the event data */
      EventNum = j & 0xFF;
      Time = (k << 8) | (j >> 8);

      if (pLink->EventFunc)
        pLink->EventFunc(pLink->Card, EventNum, Time);

      /* Schedule processing for any event-driven records */
      scanIoRequest(pLink->IoScanPvt[EventNum]);
    }
  }

  /*
   * Disable and then re-enable the IRQs so board re-issues any
   * that are still pending
   */
  pEr->Control = (status & ER_CSR_SAVE_MASK) & ~ER_CSR_INT_ENABLE;
  pEr->Control = (status & ER_CSR_SAVE_MASK) | ER_CSR_EVENT_INT;
  pEr->Control = (status & ER_CSR_SAVE_MASK) | ER_CSR_INT_ENABLE;

  status = pEr->Control; /* Flush earlier writes */
}

/******************************************************************************
 *
 * Set up the IRQs for the receiver.
 * (Should only be called once.)
 *
 ******************************************************************************/
static int ErRegisterIrq(ErLinkStruct *pLink, int IrqVector, int IrqLevel)
{
  volatile ApsErStruct *pEr = pLink->pEr;
  long status;

  if (IrqVector != -1) {
    status = devConnectInterruptVME(IrqVector, ErIrqHandler, pLink);
    if (status) {
        printf("ErRegisterIrq devConnectInterruptVME failed %lx\n", status);
        return -1;
    }

    status = devEnableInterruptLevelVME(IrqLevel);
    if (status) {
        printf("ErRegisterIrq devEnableInterruptLevelVME failed %lx\n",status);
        return -1;
    }

    pEr->IrqVector = IrqVector;
  }
  return 0;
}

/******************************************************************************
 *
 * Enable or disable the IRQs from the Event Receiver.
 *
 ******************************************************************************/
static epicsUInt16 ErEnableIrq(ApsErStruct *pParm, epicsUInt16 Mask)
{
  volatile ApsErStruct *pEr = pParm;
  epicsUInt16 status;

  if (Mask == ER_IRQ_OFF) {
    pEr->IrqEnables = 0;
    pEr->Control = (pEr->Control & ER_CSR_SAVE_MASK) & ~ER_CSR_INT_ENABLE;
  }
  else if (Mask == ER_IRQ_TELL) {
    if (pEr->Control & ER_CSR_INT_ENABLE)
      return pEr->IrqEnables & ER_IRQ_ALL;
  }
  else {
    pEr->IrqEnables = Mask;
    pEr->Control = (pEr->Control & ER_CSR_SAVE_MASK) | ER_CSR_INT_ENABLE;
  }

  status = pEr->Control; /* Flush earlier writes */
  return 0;
}

/******************************************************************************
 *
 * We load what ever RAM is not running. And Enable it.
 *
 ******************************************************************************/
static int ErUpdateRam(ApsErStruct *pParm, epicsUInt16 *RamBuf)
{
  int ChosenRam = 2;

  /* Find the idle RAM */
  if (ErGetRamStatus(pParm, 1) == 0)
    ChosenRam = 1;      /* RAM 1 is currently idle */

  ErProgramRam(pParm, RamBuf, ChosenRam);

  ErEnableRam(pParm, ChosenRam);

  return 0;
}

/******************************************************************************
 *
 * Download the entire RAM from a buffer.
 *
 ******************************************************************************/
static int ErProgramRam(ApsErStruct *pParm, epicsUInt16 *RamBuf, int Ram)
{
  volatile ApsErStruct *pEr = pParm;
  int j;
  epicsUInt16 status;

  if(ErDebug)
    printf("RAM download for RAM %d\n", Ram);

  if (Ram == 1) {
    if ((pEr->Control & ER_CSR_MAP_MASK) == ER_CSR_MAP_RAM1)
      return -1; /* RAM 1 is enabled */

    /* select VME access to ram 1 */
    pEr->Control = (pEr->Control & ER_CSR_SAVE_MASK) & ~ER_CSR_RAM_SELECT;
  }
  else if (Ram == 2) {
    if ((pEr->Control & ER_CSR_MAP_MASK) == ER_CSR_MAP_RAM2)
      return -1; /* RAM 2 is enabled */

    /* select VME access to ram 2 */
    pEr->Control = (pEr->Control & ER_CSR_SAVE_MASK) | ER_CSR_RAM_SELECT;
  }
  else
    return -1;

  /* Reset RAM address */
  pEr->Control = (pEr->Control & ER_CSR_SAVE_MASK) | ER_CSR_RAMADR_RESET;

  /* enable auto increment */
  pEr->Control = (pEr->Control & ER_CSR_SAVE_MASK) | ER_CSR_RAMADR_AUTO;

  pEr->EventRamAddr = 0;

  for (j = 0; j < NUM_EVENTS; j++) {
    if (ErDebug) {
      int k = pEr->EventRamAddr & 0xFF;

      if (k != j)
        printf("Index:%d  RamAddr:%d\n", j, k);
    }
    pEr->EventRamData = RamBuf[j];
    if (ErDebug && (RamBuf[j] != 0))
      printf("%d: %d\n", j, RamBuf[j]);
  }
  /* Turn off auto-increment */
  pEr->Control = (pEr->Control & ER_CSR_SAVE_MASK) & ~ER_CSR_RAMADR_AUTO;

  status = pEr->Control; /* Flush earlier writes */
  return 0;
}

/******************************************************************************
 *
 * Print the entire RAM from a buffer.  We skip the null events.
 *
 ******************************************************************************/
static int ErDumpRam(ApsErStruct *pParm, int Ram)
{
  volatile ApsErStruct *pEr = pParm;
  int j;

  if (Ram == 1) {
    if ((pEr->Control & ER_CSR_MAP_MASK) == ER_CSR_MAP_RAM1)
      return -1; /* RAM 1 is enabled */

    /* select VME access to ram 1 */
    pEr->Control = (pEr->Control & ER_CSR_SAVE_MASK) & ~ER_CSR_RAM_SELECT;
  }
  else if (Ram == 2) {
    if ((pEr->Control & ER_CSR_MAP_MASK) == ER_CSR_MAP_RAM2)
      return -1; /* RAM 2 is enabled */

    /* select VME access to ram 2 */
    pEr->Control = (pEr->Control & ER_CSR_SAVE_MASK) | ER_CSR_RAM_SELECT;
  }
  else
    return -1;

  /* Reset RAM address */
  pEr->Control = (pEr->Control & ER_CSR_SAVE_MASK)|ER_CSR_RAMADR_RESET;

  /* enable auto increment */
  pEr->Control = (pEr->Control & ER_CSR_SAVE_MASK)|ER_CSR_RAMADR_AUTO;

  pEr->EventRamAddr = 0;

  for (j = 0; j < NUM_EVENTS; j++) {
    int x = pEr->EventRamAddr & 0xFF;
    int y = pEr->EventRamData;

    if (y != 0)
      printf("RAM %d, %d\n", x, y);
  }

  /* turn off auto-increment */
  pEr->Control = (pEr->Control & ER_CSR_SAVE_MASK) & ~ER_CSR_RAMADR_AUTO;

  return 0;
}

/******************************************************************************
 *
 * Select the specified RAM for operation
 *
 ******************************************************************************/
static int ErEnableRam(ApsErStruct *pParm, int Ram)
{
  volatile ApsErStruct *pEr = pParm;
  epicsUInt16 status;

  if (Ram == 0) {
    pEr->Control = (pEr->Control & ER_CSR_SAVE_MASK) & ~ER_CSR_MAP_ENABLE;

    status = pEr->Control; /* Flush earlier writes */
    return 0;
  }

  if (Ram == 1)
    pEr->Control = (pEr->Control & ER_CSR_SAVE_MASK) & ~ER_CSR_MAP_SELECT;
  else if (Ram == 2)
    pEr->Control = (pEr->Control & ER_CSR_SAVE_MASK) | ER_CSR_MAP_SELECT;
  else
    return -1;

  /* Set the map enable bit */
  pEr->Control = (pEr->Control & ER_CSR_SAVE_MASK) | ER_CSR_MAP_ENABLE;

  status = pEr->Control; /* Flush earlier writes */
  return 0;
}

/******************************************************************************
 *
 * Return 1 if the requested RAM is enabled, 0 otherwise.  We return -1 on
 * error.
 *
 ******************************************************************************/
static int ErGetRamStatus(ApsErStruct *pParm, int Ram)
{
  volatile ApsErStruct *pEr = pParm;

  /* Check if RAM map is not enabled */
  if ((pEr->Control & ER_CSR_MAP_ENABLE) == 0)
    return 0;

  if (Ram == 1)
    return (pEr->Control & ER_CSR_MAP_SELECT) == 0;

  if (Ram == 2)
    return (pEr->Control & ER_CSR_MAP_SELECT) != 0;

  return -1; /* We should never get here, invalid RAM number */
}

/******************************************************************************/
/******************************************************************************/
/******************************************************************************/
/******************************************************************************/


static int DumpEventFifo(ApsErStruct *pParm)
{
  volatile ApsErStruct *pEr = pParm;

  while (pEr->Control & ER_CSR_FIFO_DATA) {
    int key;
    epicsUInt16 j, k;

    key = epicsInterruptLock();
    j = pEr->EventFifo;
    k = pEr->EventTimeHi;
    epicsInterruptUnlock(key);

    printf("FIFO event: %4.4X%4.4X\n", k, j);
  }
  return 0;
}

#ifndef vxWorks
# define ER ERchk
#endif
int ER(void)
{
  static int    Card = 0;
  char          buf[100];

  printf("Event Receiver testerizer\n");
  printf("Enter Card number [%d]: ", Card);
  fflush(stdout);
  fgets(buf, sizeof(buf), stdin);
  sscanf(buf, "%d", &Card);

  if (ErHaveReceiver(Card) < 0) {
    printf("Invalid card number %d\n", Card);
    return -1;
  }
  printf("Card address: %p\n", ErLink[Card].pEr);

  while(1) {
    printf("ER Commands:\n"
      "    r - Reset everything on the event receiver\n"
      "    f - drain the event FIFO\n"
      "    1 - dump event RAM 1\n"
      "    2 - dump event RAM 2\n"
      "    z - Dump event receiver regs\n"
      "    q - quit\n"
      "ER> "
    );
    fflush(stdout);

    if (fgets(buf, sizeof(buf), stdin) == NULL)
      break;

    switch (buf[0]) {
    case 'r':
      ErResetAll(ErLink[Card].pEr);
      break;

    case 'd':
      ErEnableRam(ErLink[Card].pEr, 0);
      break;

    case 'f':
      DumpEventFifo(ErLink[Card].pEr);
      break;

    case '1':
    case '2':
      ErDumpRam(ErLink[Card].pEr, buf[0] - '0');
      break;

    case 'z':
      ErDumpRegs(ErLink[Card].pEr);
      break;

    case 'q':
      return 0;
    }
  }
  return 0;
}

static void testErCallFunc(const iocshArgBuf *args)
{
    ER();
}
static const iocshFuncDef testErFuncDef = {"testER",0,NULL};

static void ErConfigureCallFunc(const iocshArgBuf *args)
{
    ErConfigure(args[0].ival, args[1].ival, args[2].ival, args[3].ival);
}
static const iocshArg ErConfigureArg0 = { "card number",iocshArgInt};
static const iocshArg ErConfigureArg1 = { "VME A16 address",iocshArgInt};
static const iocshArg ErConfigureArg2 = { "interrupt vector",iocshArgInt};
static const iocshArg ErConfigureArg3 = { "interrupt level",iocshArgInt};
static const iocshArg * const ErConfigureArgs[4] = {
    &ErConfigureArg0,&ErConfigureArg1,&ErConfigureArg2,&ErConfigureArg3};
static const iocshFuncDef ErConfigureFuncDef = {"ErConfigure",4,ErConfigureArgs};
static void ERRegister(void)
{
    iocshRegister(&testErFuncDef,testErCallFunc);
    iocshRegister(&ErConfigureFuncDef,ErConfigureCallFunc);
}
epicsExportRegistrar(ERRegister);
